//-----------------------------------------------------------------------------
// File:     NewtonsAppleForVisualization.cpp
// Class:    None
// Parent:   None
// Children: None
// Purpose:  Simulates A Spinning Book
//-----------------------------------------------------------------------------
// The following are standard C/C++ header files.
// If a filename is enclosed inside < >  it means the header file is in the Include directory.
// If a filename is enclosed inside " "  it means the header file is in the current directory.
#include <ctype.h>      // Character Types
#include <math.h>       // Mathematical Constants
#include <stdarg.h>     // Variable Argument Lists
#include <stdio.h>      // Standard Input/Output Functions
#include <stdlib.h>     // Utility Functions
#include <string.h>     // String Operations
#include <signal.h>     // Signals (Contol-C + Unix System Calls)
#include <setjmp.h>     // Nonlocal Goto (For Control-C)
#include <time.h>       // Time and Date information
#include <assert.h>     // Verify Program Assertion
#include <errno.h>      // Error Codes (Used in Unix system())
#include <float.h>      // Floating Point Constants
#include <limits.h>     // Implementation Constants
#include <stddef.h>     // Standard Definitions
#include <exception>    // Exception handling (e.g., try, catch throw)
//-----------------------------------------------------------------------------
#include "SimTKsimbody.h"
using namespace SimTK;
using namespace std;
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Prototypes for local functions (functions not called by code in other files)
//-----------------------------------------------------------------------------
bool  SimulateSpinningBook( void );
bool  WriteStringToFile(   const char outputString[], FILE *fptr )  { return fputs( outputString, fptr ) != 0; }
bool  WriteStringToScreen( const char outputString[] )              { return WriteStringToFile( outputString, stdout ); }
bool  WriteDoubleToFile( double x, int precision, FILE *fptr );
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute );


//-----------------------------------------------------------------------------
// The executable program starts here
//-----------------------------------------------------------------------------
int  main( int numberOfCommandLineArguments, char *arrayOfCommandLineArguments[] )
{
   // Simulate the multibody system
   bool simulationSucceeded = SimulateSpinningBook();

   // Keep the screen displayed until the user presses the Enter key
   WriteStringToScreen( "\n\n Press  Enter  to terminate the program: " );
   getchar();

   // The value returned by the main function is the exit status of the program.
   // A normal program exit returns 0 (other return values usually signal an error).
   return simulationSucceeded == true ? 0 : 1;
}



//-----------------------------------------------------------------------------
bool  SimulateSpinningBook( void )
{
   // Declare a multibody system (contains one or more force and matter sub-systems)
   MultibodySystem  mbs;

   // 0. The ground's right-handed, orthogonal x,y,z unit vectors are directed with x horizontally right and y vertically upward.
   // 1. Create a gravity vector that is straight down (in the ground's frame)
   // 2. Create a uniform gravity sub-system
   // 3. Add the gravity sub-system to the multibody system
   //Vec3 gravityVector( 0, -9.8, 0 );
   //UniformGravitySubsystem gravity( gravityVector );
   ///mbs.addForceSubsystem( gravity );

   // Create a matter sub-system (the apple)
   SimbodyMatterSubsystem  book;

   // Create the mass, center of mass, and inertia properties for the apple
   const Real  massOfBook = 0.400;

   // The location of the apple's center of mass is a vector from the apple's
   // origin expressed in the "x, y, z"' unit vectors fixed in the apple's frame.
   // Example: The vector(0,0,0) locates the apple's center of mass at the apple's origin.
   // Example: The vector(1,0,0) locates the apple's center of mass 1 unit in the "x" direction from the apple's origin.
   const Vec3  bookCenterOfMassLocation( 0, 0, 0 );

   // Create the apple's inertia matrix about its origin for the "x, y, z" unit vectors fixed in the apple's frame.
   // Note: The 3x3 inertia matrix is symmetric - so only 6 elements need to be defined.
   // Ixx, Iyy, Izz  are moments  of inertia (    diagonal terms in the matrix)
   // Ixy, Ixz, Iyz  are products of inertia (off-diagonal terms in the matrix)
   const Real  Ixx = .00269666666,  Iyy = .00169666666,   Izz = .00422666666;
   const Real  Ixy = 0,  Ixz = 0,   Iyz = 0;
   const Inertia  bookInertiaMatrix( Ixx, Iyy, Izz, Ixy, Ixz, Iyz );

   // The MassProperties class holds the mass, center of mass, and inertia properties of a rigid body.
   // Although the next line creates an instance of the MassProperties class for the apple,
   // it does not get associated with the apple until the  addRigidBody  method.
   MassProperties  bookMassProperties( massOfBook, bookCenterOfMassLocation, bookInertiaMatrix );

   // The motion of the apple is related to the motion of the ground via "mobilizers"
   // The "mobilizers" specify the allowable motion of the apple to the ground.
   // More specifically, the motion of an "outboard body" (e.g., the apple)
   // to its inboard body (e.g., the ground) is specified by first constructing:
   // 1. An  "outboard frame"  on the ground (which hooks to the "outboard body" - the apple )
   // 2. An  "inboard  frame"  on the apple  (which hooks to the "inboard  body" - the ground)

   // The orientation and position of the  outboard frame  from the ground's frame is specified below.
   // The outboard frame's axes are aligned with the ground's axes and its origin is coincident with the ground's origin.
   // In other words, for this simple problem the outboard frame and the ground frame are identical.
   const Transform outboardFrameTransformFromGround;  // The default constructor is the identity transform

   // The orientation and position of the  inboard  frame  from the apple's  frame is specified below.
   // The inboard frame's  axes are aligned with the apple's  axes and its origin is coincident with the apple's  origin
   // In other words, for this simple problem the inboard  frame and the apple frame are identical.
   // Although the inboard frame can be constructed in a simple manner analogous to the outboardFrameTransformFromGround (above)
   // it is worthwhile to look at the details of the rotation matrix and position vector in the transform:
   // a. The rotation matrix relating the InboardFrame's x,y,z axes to the AppleFrame's x,y,z axes is specified InboardFrame_AppleFrame.
   // b. The position of the InboardFrame's origin from the AppleFrame origin, expressed in terms of the AppleFrame's "x, y, z" unit vectors.
   const Rotation   inboardFrameOrientationInBook;   // ( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3       inboardFrameOriginLocationFromBookOrigin( 0, 0, 0 );
   const Transform  inboardFrameTransformFromBook( inboardFrameOrientationInBook, inboardFrameOriginLocationFromBookOrigin );

   // There are many ways that the apple can move relative to the ground.
   // The following allows the apple to move in "x", "y", and "z" directions.
   // Another option producing the same result is Mobilizer::Free
   Mobilizer bookToGroundMobilizer = Mobilizer::Free;
   const BodyId bookBodyId = book.addRigidBody( bookMassProperties, inboardFrameTransformFromBook, GroundId, outboardFrameTransformFromGround, bookToGroundMobilizer );

   // Add the matter (apple) sub-system to the system.
   mbs.setMatterSubsystem( book );

   // Create a state for this system.
   // Define appropriate states for this multi-body system.
   // Set the initial time to 0.0
   State s;
   mbs.realize( s );
   s.setTime( 0.0 );

   // Set the initial values for the configuration variables (e0,e1,e2,e3,x,y,z)
   // Expressed in Body Frame but this is the Same as Ground Initially
   book.setMobilizerQ( s, bookBodyId, 0,  1.0 );
   book.setMobilizerQ( s, bookBodyId, 1,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 2,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 3,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 4,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 5,  0.0 );
   book.setMobilizerQ( s, bookBodyId, 6,  0.0 );
   
   // Set the initial values for the motion variables using the given information
   const Real wx = 7.0, wy = 0.2, wz = 0.2;
   //const Real wx = 0.2, wy = 7.0, wz = 0.2;
   //const Real wx = 0.2, wy = 0.2, wz = 7.0;

   // Set the initial values for the motion variables
   // Expressed in the Body Frame (0-2) Angular Velocities, (3-5) Linear Velocities
   book.setMobilizerU( s, bookBodyId, 0,  wx  );
   book.setMobilizerU( s, bookBodyId, 1,  wy  );
   book.setMobilizerU( s, bookBodyId, 2,  wz  );
   book.setMobilizerU( s, bookBodyId, 3,  0.0 );
   book.setMobilizerU( s, bookBodyId, 4,  0.0 );
   book.setMobilizerU( s, bookBodyId, 5,  0.0 ); 
     
   // Create a study using the Runge Kutta Merson integrator (alternately use the CPodesIntegrator)
   RungeKuttaMerson myStudy( mbs, s );

   // Set the numerical accuracy for the integrator
   myStudy.setAccuracy( 1.0E-7 );

   // The next statement does lots of accounting
   myStudy.initialize();

   // Open a file to record the simulation results (they are also displayed on screen)
   FILE *outputFile = FileOpenWithMessageIfCannotOpen( "SpinningBookForPlottingResults.txt", "w" );
   WriteStringToFile(   "time     wx      wy      wz     K      Hx      Hy      Hz      Hmag\n",  outputFile );
   WriteStringToScreen( "time     wx      wy      wz     K      Hx      Hy      Hz      Hmag\n" );

   // Visualize results with VTKReporter
   VTKReporter animationResults( mbs );

   // For visualization purposes only, create a red book with specified half heights
   const Real  height = .22, width = .28, thickness = .05;
   const Vec3 bookHalfHeights(height/2, width/2, thickness/2);
   DecorativeBrick redBook = DecorativeBrick(bookHalfHeights);
   redBook.setColor(Red);     // Can also specify a Vec3 with rgb which scale from 0 to 1
   redBook.setOpacity(0.4);   // 0.0 is solid and 1.0 is transparent
   
   // Decorate the apple with the red sphere at the apple's origin
   const Rotation  bookToRedSphereOrientation; //( 1,0,0, 0,1,0, 0,0,1 );
   const Vec3      bookOriginToRedSphereOriginLocation( 0, 0, 0 );
   const Transform bookToRedSphereTransform( bookToRedSphereOrientation, bookOriginToRedSphereOriginLocation );
   animationResults.addDecoration( bookBodyId, bookToRedSphereTransform, redBook );

   // Set the numerical integration step and the time for the simulation to run
   const Real dt = 0.01;
   const Real finalTime = 4.0;

   // Run the simulation and print the results
   while( 1 )
   {
      // Query for results to be printed
      Real time = s.getTime();
      Real kineticEnergy = mbs.getKineticEnergy(s);
      Real uniformGravitationalPotentialEnergy = mbs.getPotentialEnergy(s);
      Real mechanicalEnergy = kineticEnergy + uniformGravitationalPotentialEnergy;

      // Retrieve the Angular Velocity Components of the Book in the Ground
      Vec3 bookAngularVelocityinGround = book.calcBodyAngularVelocityInBody( s, bookBodyId, GroundId );
      
      // Retrieve the Rotation Matrix from the Ground to the Book
      // .calcBodyRotationFromBody(state S, bodyA, bodyB) returns R_B_A
      Rotation groundToBookRotationMatrix = book.calcBodyRotationFromBody( s, GroundId, bookBodyId);
      
      // Calculate the Angular Velocity in the book Frame
      Vec3 bookAngularVelocityinBody = groundToBookRotationMatrix * bookAngularVelocityinGround;
      Real xAngularVelocity = bookAngularVelocityinBody[0];
      Real yAngularVelocity = bookAngularVelocityinBody[1];
      Real zAngularVelocity = bookAngularVelocityinBody[2];

      // Retreive the Angular Momentum about the Book Center of Mass in the Ground Frame
      SpatialVec bookBodyMomentumAboutCMInGround = book.calcBodyMomentumAboutBodyMassCenterInGround(s, bookBodyId);
      //  Spatial Vec is Two Vec3 - Top is AngularMomenum, Lower is LinearMomentum
      Vec3 bookAngularMomentumAboutCMInGround = bookBodyMomentumAboutCMInGround[0];

      // Calculate the Angular Momentum Components in the Book Frame;
      Vec3 bookAngularMomentumAboutCMInBook = groundToBookRotationMatrix * bookAngularMomentumAboutCMInGround;
      Real xAngularMomentum = bookAngularMomentumAboutCMInBook[0];
      Real yAngularMomentum = bookAngularMomentumAboutCMInBook[1];
      Real zAngularMomentum = bookAngularMomentumAboutCMInBook[2];
      Real magAngularMomentum = pow( ( pow(xAngularMomentum,2) + pow(yAngularMomentum,2)+ pow(zAngularMomentum,2) ), .5);
      
      // Print results to screen
      WriteDoubleToFile( time,                 2, stdout );
      WriteDoubleToFile( xAngularVelocity,     4, stdout );
      WriteDoubleToFile( yAngularVelocity,     4, stdout );
      WriteDoubleToFile( zAngularVelocity,     4, stdout );
      WriteDoubleToFile( mechanicalEnergy,     4, stdout );
      WriteDoubleToFile( xAngularMomentum,     4, stdout );
      WriteDoubleToFile( yAngularMomentum,     4, stdout );
      WriteDoubleToFile( zAngularMomentum,     4, stdout );
      WriteDoubleToFile( magAngularMomentum,   4, stdout );
      WriteStringToScreen( "\n" );

      // Print results to file
      WriteDoubleToFile( time,                 2, outputFile );
      WriteDoubleToFile( xAngularVelocity,     4, outputFile );
      WriteDoubleToFile( yAngularVelocity,     4, outputFile );
      WriteDoubleToFile( zAngularVelocity,     4, outputFile );
      WriteDoubleToFile( mechanicalEnergy,     4, outputFile );
      WriteDoubleToFile( xAngularMomentum,     4, outputFile );
      WriteDoubleToFile( yAngularMomentum,     4, outputFile );
      WriteDoubleToFile( zAngularMomentum,     4, outputFile );
      WriteDoubleToFile( magAngularMomentum,   4, outputFile );
      WriteStringToFile( "\n", outputFile );

      // Animate the results for this step
      animationResults.report(s);

      // Check if integration has completed
      if( time >= finalTime ) break;

      // Increment time step
      myStudy.step( time + dt);
   }

   // Simulation completed properly
   return true;
}


//-----------------------------------------------------------------------------
FILE*  FileOpenWithMessageIfCannotOpen( const char *filename, const char *attribute )
{
   // Try to open the file
   FILE *Fptr1 = fopen( filename, attribute );

   // If unable to open the file, issue a message
   if( !Fptr1 )
   {
      WriteStringToScreen( "\n\n Unable to open the file: " );
      WriteStringToScreen( filename );
      WriteStringToScreen( "\n\n" );
   }

   return Fptr1;
}


//-----------------------------------------------------------------------------
bool  WriteDoubleToFile( double x, int precision, FILE *fptr )
{
   // Ensure the precision (number of digits in the mantissa after the decimal point) makes sense.
   // Next, calculate the field width so it includes one extra space to the right of the number.
   if( precision < 0 || precision > 17 ) precision = 5;
   int fieldWidth = precision + 8;

   // Create the format specifier and print the number
   char format[20];
   sprintf( format, " %%- %d.%dE", fieldWidth, precision );
   return fprintf( fptr, format, x ) >= 0;
}


